{$A+,B-,D+,E+,F-,G+,I+,L+,N+,P+,Q-,R-,S-,T-,V+,X+,Y+}

(* CANVAS-EFFECTS
	 --------------
(C) 2000 by Arsne von Wyss - arsene@vonwyss.ch

Rev: 20000214

This source code is freeware for non-commercial use and provided AS-IS.
You are not allowed to use or spread incomplete (inclusive this header)
copies of this code. Use at your own risk. If you do not agree with these
statements, delete this source code immediately. Please contact the
author via email for questions or bug reports.

Source code in development and subject to change frequently. No support from
the author yet.*)

unit CanvasFX;

interface

uses
	Objects;

type
	TColor=Longint;
	PBitmapProcessor=^TBitmapProcessor;
	TBitmapProcessor=object(TObject)
		function MemorySize(Width: Integer; Fmt: Word): Word; virtual;
		procedure Convert(const Src; SrcFmt: Word; var Dst; DstFmt: Word; Width: Integer); virtual;
		procedure AlphaFill(var Dst; Fmt: Word; Width: Integer; Col: TColor); virtual;
		procedure AlphaMix(const Src; var Dst; Fmt: Word; Width: Integer; Alpha: Byte); virtual;
		procedure EqualMix(const Src; var Dst; Fmt: Word; Width: Integer); virtual;
		procedure ConvertToScreen(const Src; SrcFmt: Word; var Dst; Width: Integer);
		procedure ConvertFromScreen(const Src; var Dst; DstFmt: Word; Width: Integer);
		procedure AlphaFillScreen(var Dst; Width: Integer; Col: TColor);
		procedure AlphaMixScreen(const Src; var Dst; Width: Integer; Alpha: Byte);
	end;
	TColorCode=object
		Code: Word;
		function Bits: Word;
		function Bytes: Word;
	end;
	TRGBColor=packed record
		case Boolean of
		False: (RGB: TColor);
		True: (R,G,B,A: Byte);
	end;

const
	bfMono=$1000;
	bf8=$0222;
	bf15=$555;
	bf16=$565;
	bf24=$888;
	bf32=$8888;
	alphaEqual=$80;

var
	BitmapProcessor: PBitmapProcessor;

procedure CreatePalette(var Pal);
function RGBToPalette(RGB: Longint): Byte;
function PaletteToRGB(Col: Byte): Longint;

implementation

uses
	GUI; {needs TScreen.BitmapFormat}

{8 Bit format
 ------------
2 Bits Red
2 Bits Green
2 Bits Blue
= 64 Base Colors

2 Bits Luminosity
= 256 Total Colors}

function TBitmapProcessor.MemorySize(Width: Integer; Fmt: Word): Word;
begin
	MemorySize:=TColorCode(Fmt).Bytes*Width;
end;

procedure TBitmapProcessor.Convert(const Src; SrcFmt: Word; var Dst; DstFmt: Word; Width: Integer);
begin
	if (SrcFmt=DstFmt) or (Width<=0) then
		Exit;
	case SrcFmt of {Mono support not done yet}
{	bfMono: case DstFmt of
		bf8: asm
			end;
		bf15: asm
			end;
		bf16: asm
			end;
		bf24: asm
			end;
		bf32: asm
			end;
		else
			RunError;
		end;}
	bf8: case DstFmt of
{		bfMono: asm
			end;}
		bf15: asm
				push ds
				lds si,Src
				les di,Dst
				cld
			@loop:
				lodsb
				push ax
				call PaletteToRGB {AL=R,AH=G,DL=B}
				mov bx,ax
				shr ax,3
				and bx,0000000000011111b
				shr bx,6
				and bx,0000001111100000b
				shl dx,10
				and bx,0111110000000000b
				or ax,bx
				or ax,dx
				stosw
				dec Width
				jnz @loop
				pop ds
			end;
		bf16: asm
				push ds
				lds si,Src
				les di,Dst
				cld
			@loop:
				lodsb
				push ax
				call PaletteToRGB {AL=R,AH=G,DL=B}
				mov bx,ax
				shr ax,3
				and bx,0000000000011111b
				shr bx,5
				and bx,0000011111100000b
				shl dx,11
				and bx,1111100000000000b
				or ax,bx
				or ax,dx
				stosw
				dec Width
				jnz @loop
				pop ds
			end;
		bf24: asm
				push ds
				lds si,Src
				les di,Dst
				cld
			@loop:
				lodsb
				push ax
				call PaletteToRGB {AL=R,AH=G,DL=B}
				stosw
				mov es:[di],dl
				inc di
				dec Width
				jnz @loop
				pop ds
			end;
		bf32: asm
				push ds
				lds si,Src
				les di,Dst
				cld
			@loop:
				lodsb
				push ax
				call PaletteToRGB {AL=R,AH=G,DL=B}
				stosw
				mov es:[di],dx
				add di,2
				dec Width
				jnz @loop
				pop ds
			end;
		else
			RunError;
		end;
	bf15: case DstFmt of
{		bfMono: asm
			end;}
		bf8: asm
				push ds
				lds si,Src
				les di,Dst
				cld
			@loop:
				lodsw
				mov dx,ax
				shl ax,3
				shl ah,3
				shr dx,7
				and dx,0000000011111000b
				push dx
				push ax
				call RGBToPalette
				stosb
				dec Width
				jnz @loop
				pop ds
			end;
		bf16: asm
				push ds
				lds si,Src
				les di,Dst
				cld
			@loop:
				lodsw
				mov bx,ax
				and ax,0000000000011111b
				shl bx,1
				and bx,1111111111000000b
				or ax,bx
				stosw
				dec Width
				jnz @loop
				pop ds
			end;
		bf24: asm
				push ds
				lds si,Src
				les di,Dst
				cld
			@loop:
				lodsw
				mov dx,ax
				shl ax,3
				shl ah,3
				shr dx,7
				and dx,0000000011111000b
				stosw
				mov es:[di],dl
				inc di
				dec Width
				jnz @loop
				pop ds
			end;
		bf32: asm
				push ds
				lds si,Src
				les di,Dst
				cld
			@loop:
				lodsw
				mov dx,ax
				shl ax,3
				shl ah,3
				shr dx,7
				and dx,0000000011111000b
				stosw
				mov es:[di],dx
				add di,2
				dec Width
				jnz @loop
				pop ds
			end;
		else
			RunError;
		end;
	bf16: case DstFmt of
{		bfMono: asm
			end;}
		bf8: asm
				push ds
				lds si,Src
				les di,Dst
				cld
			@loop:
				lodsw
				mov dx,ax
				shl ax,3
				shl ah,2
				shr dx,8
				and dx,0000000011111000b
				push dx
				push ax
				call RGBToPalette
				stosb
				dec Width
				jnz @loop
				pop ds
			end;
		bf15: asm
				push ds
				lds si,Src
				les di,Dst
				cld
			@loop:
				lodsw
				mov bx,ax
				and ax,0000000000011111b
				shr bx,1
				and bx,0111111111100000b
				or ax,bx
				stosw
				dec Width
				jnz @loop
				pop ds
			end;
		bf24: asm
				push ds
				lds si,Src
				les di,Dst
				cld
			@loop:
				lodsw
				mov dx,ax
				shl ax,3
				shl ah,2
				shr dx,8
				and dx,0000000011111000b
				stosw
				mov es:[di],dl
				inc di
				dec Width
				jnz @loop
				pop ds
			end;
		bf32: asm
				push ds
				lds si,Src
				les di,Dst
				cld
			@loop:
				lodsw
				mov dx,ax
				shl ax,3
				shl ah,2
				shr dx,8
				and dx,0000000011111000b
				stosw
				mov es:[di],dx
				add di,2
				dec Width
				jnz @loop
				pop ds
			end;
		else
			RunError;
		end;
	bf24: case DstFmt of
{		bfMono: asm
			end;}
		bf8: asm
				push ds
				lds si,Src
				les di,Dst
				cld
			@loop:
				mov al,ds:[si+2]
				push ax
				push word ptr ds:[si]
				add si,3
				call RGBToPalette
				stosb
				dec Width
				jnz @loop
				pop ds
			end;
		bf15: asm
				push ds
				lds si,Src
				les di,Dst
				cld
			@loop:
				lodsw
				shr ah,3
				shr ax,3
				mov bh,ds:[si]
				inc si
				shr bh,1
				and bx,0111110000000000b
				or ax,bx
				stosw
				dec Width
				jnz @loop
				pop ds
			end;
		bf16: asm
				push ds
				lds si,Src
				les di,Dst
				cld
			@loop:
				lodsw
				shr ah,2
				shr ax,3
				mov bh,ds:[si]
				inc si
				and bx,1111100000000000b
				or ax,bx
				stosw
				dec Width
				jnz @loop
				pop ds
			end;
		bf32: asm
				push ds
				lds si,Src
				les di,Dst
				cld
				sub al,al
			@loop:
				movsw
				movsb
				stosb
				dec Width
				jnz @loop
				pop ds
			end;
		else
			RunError;
		end;
	bf32: case DstFmt of
{		bfMono: asm
			end;}
		bf8: asm
				push ds
				lds si,Src
				les di,Dst
				cld
			@loop:
				push word ptr ds:[si+2]
				push word ptr ds:[si]
				add si,4
				call RGBToPalette
				stosb
				dec Width
				jnz @loop
				pop ds
			end;
		bf15: asm
				push ds
				lds si,Src
				les di,Dst
				cld
			@loop:
				lodsw
				shr ah,3
				shr ax,3
				mov bh,ds:[si]
				add si,2
				shr bh,1
				and bx,0111110000000000b
				or ax,bx
				stosw
				dec Width
				jnz @loop
				pop ds
			end;
		bf16: asm
				push ds
				lds si,Src
				les di,Dst
				cld
			@loop:
				lodsw
				shr ah,2
				shr ax,3
				mov bh,ds:[si]
				add si,2
				and bx,1111100000000000b
				or ax,bx
				stosw
				dec Width
				jnz @loop
				pop ds
			end;
		bf24: asm
				push ds
				lds si,Src
				les di,Dst
				cld
			@loop:
				movsw
				movsb
				inc si
				dec Width
				jnz @loop
				pop ds
			end;
		else
			RunError;
		end;
	else
		RunError;
	end;
end;

procedure TBitmapProcessor.AlphaFill(var Dst; Fmt: Word; Width: Integer; Col: TColor);
var
	RGBA: packed record
		R,G,B,A: Byte;
	end absolute Col;
begin
	if (Width<=0) or (RGBA.A=255) then
		Exit;
	asm
		mov bl,$FF
		sub bl,RGBA.A
		mov al,RGBA.R
		mul bl
		mov RGBA.R,ah
		mov al,RGBA.G
		mul bl
		mov RGBA.G,ah
		mov al,RGBA.B
		mul bl
		mov RGBA.B,ah
	end;
	case Fmt of
	bf8: asm
			les di,Dst
			cld
		@loop:
			mov al,es:[di]
			push ax
			call PaletteToRGB
			mov cl,ah
			mul RGBA.A
			add ah,RGBA.R
			xchg cx,ax {CH: R, AL: G}
			mul RGBA.A
			add ah,RGBA.G
			xchg dx,ax {DH: G, AL: B}
			mul RGBA.A
			add ah,RGBA.B
			shr ax,8
			push ax
			mov dl,ch
			push dx
			call RGBToPalette
			stosb
			dec Width
			jnz @loop
		end;
	bf15: asm
			shr RGBA.R,3
			shr RGBA.G,3
			shr RGBA.B,3
			les di,Dst
			cld
		@loop:
			mov ax,es:[di]
			mov bx,ax
			and al,0000000000011111b
			mul RGBA.A
			add ah,RGBA.R
			mov cl,ah
			sub ch,ch
			mov ax,bx
			shr ax,5
			and al,0000000000011111b
			mul RGBA.A
			add ah,RGBA.G
			sub al,al
			shr ax,3
			or cx,ax
			mov ax,bx
			shr ax,10
			and al,0000000000011111b
			mul RGBA.A
			add ah,RGBA.B
			shl ax,2
			and ax,0111110000000000b
			or ax,cx
			stosw
			dec Width
			jnz @loop
		end;
	bf16: asm
			shr RGBA.R,3
			shr RGBA.G,2
			shr RGBA.B,3
			les di,Dst
			cld
		@loop:
			mov ax,es:[di]
			mov bx,ax
			and al,0000000000011111b
			mul RGBA.A
			add ah,RGBA.R
			mov cl,ah
			sub ch,ch
			mov ax,bx
			shr ax,5
			and al,0000000000111111b
			mul RGBA.A
			add ah,RGBA.G
			sub al,al
			shr ax,3
			or cx,ax
			mov ax,bx
			shr ax,11
			mul RGBA.A
			add ah,RGBA.B
			shl ax,3
			and ax,1111100000000000b
			or ax,cx
			stosw
			dec Width
			jnz @loop
		end;
	bf24: asm
			les di,Dst
		@loop:
			mov al,es:[di]
			mul RGBA.A
			add ah,RGBA.R
			mov es:[di],ah
			inc di
			mov al,es:[di]
			mul RGBA.A
			add ah,RGBA.G
			mov es:[di],ah
			inc di
			mov al,es:[di]
			mul RGBA.A
			add ah,RGBA.B
			mov es:[di],ah
			inc di
			dec Width
			jnz @loop
		end;
	bf32: asm
			les di,Dst
		@loop:
			mov al,es:[di]
			mul RGBA.A
			add ah,RGBA.R
			mov es:[di],ah
			inc di
			mov al,es:[di]
			mul RGBA.A
			add ah,RGBA.G
			mov es:[di],ah
			inc di
			mov al,es:[di]
			mul RGBA.A
			add ah,RGBA.B
			mov es:[di],ah
			add di,2
			dec Width
			jnz @loop
		end;
	else
		RunError;
	end;
end;

procedure TBitmapProcessor.AlphaMix(const Src; var Dst; Fmt: Word; Width: Integer; Alpha: Byte);
begin
	case Fmt of
	bf8: asm
		end;
	bf15: asm
		end;
	bf16: asm
		end;
	bf24: asm
		end;
	bf32: asm
		end;
	else
		RunError;
	end;
end;

procedure TBitmapProcessor.EqualMix(const Src; var Dst; Fmt: Word; Width: Integer);
begin
	AlphaMix(Src,Dst,Fmt,Width,alphaEqual);
end;

procedure TBitmapProcessor.ConvertToScreen(const Src; SrcFmt: Word; var Dst; Width: Integer);
begin
	Convert(Src,SrcFmt,Dst,Screen^.BitmapFormat,Width);
end;

procedure TBitmapProcessor.ConvertFromScreen(const Src; var Dst; DstFmt: Word; Width: Integer);
begin
	Convert(Src,Screen^.BitmapFormat,Dst,DstFmt,Width);
end;

procedure TBitmapProcessor.AlphaFillScreen(var Dst; Width: Integer; Col: TColor);
begin
	AlphaFill(Dst,Screen^.BitmapFormat,Width,Col);
end;

procedure TBitmapProcessor.AlphaMixScreen(const Src; var Dst; Width: Integer; Alpha: Byte);
begin
	if Alpha=alphaEqual then
		EqualMix(Src,Dst,Screen^.BitmapFormat,Width)
	else
		AlphaMix(Src,Dst,Screen^.BitmapFormat,Width,Alpha);
end;

function TColorCode.Bits: Word; assembler;
asm
	les di,self
	mov ax,es:[di]
	mov bx,ax
	shr bx,4
	and ax,$0F0F
	and bl,$0F
	add ax,bx
	add al,ah
	sub ah,ah
end;

function TColorCode.Bytes: Word; assembler;
asm
	les di,self
	mov ax,es:[di]
	mov bx,ax
	shr bx,4
	and ax,$0F0F
	and bl,$0F
	add ax,bx
	add al,ah
	sub ah,ah
	add ax,7
	shr ax,3
end;

procedure CreatePalette(var Pal);
var
	Palette: packed array[Byte] of packed record
		R,G,B: Byte;
	end absolute Pal;
	I: Integer;
	Col: packed record
		case Boolean of
		True: (RGB: Longint);
		False: (R,G,B,A: Byte);
	end;
begin
	for I:=0 to 255 do begin
		Col.RGB:=PaletteToRGB(I);
		with Palette[I] do begin
			R:=Col.R;
			G:=Col.G;
			B:=Col.B;
		end;
	end;
end;

function RGBToPalette(RGB: Longint): Byte; assembler; {does not modify ES,DI,SI,Direction-Flag}
var
	Divider: Word;
asm
	mov Divider,272*16
	mov ah,byte ptr RGB+2
	sub dx,dx
	div Divider
	mov cx,ax
	mov ah,byte ptr RGB+1
	sub dx,dx
	div Divider
	mov bx,ax
	mov ah,byte ptr RGB
	sub dx,dx
	div Divider
	mov dl,al
	and dl,$3
	mov dh,bl
	and dh,$3
	add dl,dh
	add dl,dh
	mov dh,cl
	and dh,$3
	add dl,dh
	sub dh,dh
	shr dx,2
	and al,$0C
	and bl,$0C
	shl bl,2
	and cl,$0C
	shl cl,4
	or ax,bx
	or ax,cx
	or ax,dx
end;

function PaletteToRGB(Col: Byte): Longint; assembler; {does not modify ES,DI,SI,Direction-Flag}
var
	Divider: Word;
asm
	mov al,col
	mov bx,ax
	mov cx,ax
	mov dx,ax
	and dx,$03
	shl dx,4
	and ax,$C0
	add ax,dx
	and bx,$30
	shl bx,2
	add bx,dx
	and cx,$0C
	shl cx,4
	add cx,dx
	mov Divider,272
	mul Divider
	sub ax,dx
	xchg ax,bx
	mul Divider
	sub ax,dx
	xchg ax,cx
	mul Divider
	sub ax,dx
	mov dl,ah
	mov ah,ch
	mov al,bh
end;

end.