{$A+,B-,D+,E+,F-,G+,I+,L+,N+,P+,Q-,R-,S-,T-,V+,X+,Y+}

(* HWINTOBJ
	 --------
(C) 1997,2000 by Arsne von Wyss - arsene@vonwyss.ch

Rev: 20000214

This source code is freeware for non-commercial use and provided AS-IS.
You are not allowed to use or spread incomplete (inclusive this header)
copies of this code. Use at your own risk. If you do not agree with these
statements, delete this source code immediately. Please contact the
author via email for questions or bug reports.

Source code in development and subject to change frequently. No support from
the author yet.*)

unit HWIntObj; {HardWare Interrupt Object}

interface

uses Objects,DOS;

type
	TIRQ=0..15;
	PInterrupt=^TInterrupt;
	TInterrupt=object(TObject)
	private
		FNext,FPrev: PInterrupt;
		FIRQ: TIRQ;
		FBase,FMask,FSoftInt: Word;
		FWasEnabled: Boolean;
		FHandler,FOldHandler: Pointer;
	public
		constructor Init(AIRQ: TIRQ);
		destructor Done; virtual;
		procedure Handler; virtual; {CAUTION: Must be the 2nd "VIRTUAL" position!}
		function IRQ: Integer;
		procedure Enable; virtual;
		procedure Disable; virtual;
	end;

procedure CLI; inline($FA);
procedure STI; inline($FB);

implementation

var
	OldExit: Pointer;
	Ints: PInterrupt;

{!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!}
{!!! CAUTION !!! CAUTION !!! CAUTION !!! CAUTION !!! CAUTION !!! CAUTION !!!}
{!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!}
{!}
{!}{The following type declaration is a offset mask for the following pseudo-}
{!}{procedure. The procedure is never called directly but mutations of it are}
{!}{created by CreateHandler and modified for object handler call.}
{!}
{!}{If you need to modify anything in this code, so you'll have to count the}
{!}{used bytes one by one and correct the TIGData structure, else the whole}
{!}{handler system may crash on the next interrupt.}
{!}
{!}{The IGData procedure saves the registers (32 bit), pushes the object}
{!}{address on the stack and calls the "Handler" procedure retrieved from}
{!}{the objects VMT. At the very end it "free"s the interrupt controller.}
{!}
{!}type
{!}	PIGData=^TIGData;
{!}	TIGData=record
{!}		Code1: array[1..22] of Byte;
{!}		Ptr: Pointer;  { 4}
{!}		Code2: array[1.. 1] of Byte;
{!}		Vmt: Pointer;  { 4}
{!}		Code3: array[1..16] of Byte;
{!}		Adr: Word;     { 2}
{!}		Code4: array[1.. 5] of Byte;
{!}	end; {Total size: 54 Bytes}
{!}
{!}procedure IGData; far; assembler;
{!}asm
{!}	db $66; push ax
{!}	db $66; push bx
{!}	db $66; push cx
{!}	db $66; push dx
{!}	db $66; push di
{!}	db $66; push si
{!} db $66; push bp
{!}	push es
{!}	push ds
{!}	push seg @data
{!}	pop ds
{!}	db $66; push $AAAA; dw $AAAA { $AAAAAAAA => replaced by Int}
{!}	db $9A; dd $AAAAAAAA         { $AAAAAAAA => replaced by VMT}
{!}	pop ds
{!}	pop es
{!} db $66; pop bp
{!}	db $66; pop si
{!}	db $66; pop di
{!}	db $66; pop dx
{!}	db $66; pop cx
{!}	db $66; pop bx
{!}	mov al,$20
{!}	out $A0,al                   { out $A0,al => may be killed}
{!}	out $20,al
{!}	db $66; pop ax
{!}	iret
{!}end;
{!}
{!}function CreateHandler(Int: PInterrupt; PIC2: Boolean): Pointer;
{!}var
{!}	Size: Word;
{!}	Buf: PIGData;
{!} VMTBuf: Pointer;
{!}begin
{!} asm
{!} 	les bx,Int
{!} 	mov bx,es:[bx]                                  {ds:[bx] => VMT}
{!}		db $66; mov ax,word ptr ds:[bx+4+vmtHeaderSize] {?.Handler proc address}
{!}   db $66; mov word ptr VMTBuf,ax
{!} end;
{!}	New(Buf);
{!}	Move(@IGData^,Buf^,Sizeof(TIGData));
{!}	Buf^.Ptr:=Int;
{!} Buf^.VMT:=VMTBuf;
{!} if not PIC2 then
{!}		Move(Buf^.Code4,Buf^.Adr,SizeOf(Buf^.Code4));
{!}	CreateHandler:=Buf;
{!}end;
{!}
{!}procedure DestroyHandler(Handler: PIGData);
{!}begin
{!}	Dispose(Handler);
{!}end;
{!}
{!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!}
{!!! CAUTION !!! CAUTION !!! CAUTION !!! CAUTION !!! CAUTION !!! CAUTION !!!}
{!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!}

procedure NewExit; far;
begin
	ExitProc:=OldExit;
	while Ints<>nil do
		Ints^.Done;
end;

constructor TInterrupt.Init(AIRQ: TIRQ);
begin
	inherited Init;
	FSoftInt:=AIRQ+$08;
	FBase:=$20;
	FMask:=1 shl (AIRQ and 7);
	if AIRQ>=8 then begin
		Inc(FSoftInt,$60);
		Inc(FBase,$80);
	end;
	FIRQ:=AIRQ;
	FWasEnabled:=Port[FBase+1] and FMask=0;
	FHandler:=CreateHandler(@self,FBase>$20);
	if AIRQ>1 then {Do not disable timer and keyboard interrupts!}
		Disable;
	CLI;
	GetIntVec(FSoftInt,FOldHandler);
	SetIntVec(FSoftInt,FHandler);
	STI;
	FNext:=Ints;
	Ints:=@self;
	if FNext<>nil then
		FNext^.FPrev:=@self;
	Port[$20]:=$20;
	if FBase>$20 then
		Port[FBase]:=$20;
end;

destructor TInterrupt.Done;
begin
	CLI;
	SetIntVec(FSoftInt,FOldHandler);
	if FWasEnabled or (FIRQ=0) then
		Enable
	else
		Disable;
	Port[$20]:=$20;
	if FBase>$20 then
		Port[FBase]:=$20;
	STI;
	DestroyHandler(FHandler);
	if FPrev<>nil then
		FPrev^.FNext:=FNext
	else
		Ints:=FNext;
	if FNext<>nil then
		FNext^.FPrev:=FPrev;
	inherited Done;
end;

function TInterrupt.IRQ: Integer;
begin
	IRQ:=FIRQ;
end;

procedure TInterrupt.Enable;
begin
	CLI;
	Port[FBase+1]:=Port[FBase+1] and (not FMask);
	Port[$20]:=$20;
	if FBase>$20 then
		Port[FBase]:=$20;
	STI;
end;

procedure TInterrupt.Disable;
begin
	CLI;
	Port[FBase+1]:=Port[FBase+1] or FMask;
	Port[$20]:=$20;
	if FBase>$20 then
		Port[FBase]:=$20;
	STI;
end;

procedure TInterrupt.Handler; assembler;
asm
	les	di,self
	pushf
	call es:[di].FOldHandler
end;

begin
	OldExit:=ExitProc;
	ExitProc:=@NewExit;
end.
