{$A+,B-,D+,E+,F-,G+,I+,L+,N+,P+,Q-,R-,S-,T-,V+,X+,Y+}

(* MENUS
	 -----
(C) 1998,99 by Arsne von Wyss - arsene@vonwyss.ch

Rev: 19990925

This source code is freeware for non-commercial use and provided AS-IS.
You are not allowed to use or spread incomplete (inclusive this header)
copies of this code. Use at your own risk. If you do not agree with these
statements, delete this source code immediately. Please contact the
author via email for questions or bug reports. *)

{Menus: menu bars, popup menus}

unit Menus;

(* Unit interface is HelpGen enabled *)

interface

uses
	Objects,GUI,Elements,Canvas,Events;

type
	PGenericMenuItem=^TGenericMenuItem;
	TGenericMenuItem=object(TListItem)
		Caption: PString;
		Icon: PBitmap;
		Enabled: Boolean;
		function Height: Integer; virtual;
		function Width: Integer; virtual;
		procedure Setup(AOwner: PListView); virtual;
		procedure SetCaption(const ACaption: string);
		procedure Draw(Active,Focus: Boolean); virtual;
		procedure Action(Pressed: Boolean); virtual;
		procedure Close; virtual;
		destructor Done; virtual;
	end;
	PMenu=^TMenu;
	PMenuItem=^TMenuItem;
	TMenuItem=object(TGenericMenuItem)
		Command: Integer;
		procedure Action(Pressed: Boolean); virtual;
	end;
	PMenuPopup=^TMenuPopup;
	TMenuPopup=object(TGenericMenuItem)
		Menu: PMenu;
		constructor Init;
		procedure Action(Pressed: Boolean); virtual;
		procedure Draw(Active,Focus: Boolean); virtual;
		procedure Close; virtual;
		destructor Done; virtual;
	end;
	PMenuList=^TMenuList;
	TMenuList=object(TListView)
		procedure HandleEvent(var Event: TEvent); virtual;
	end;
	PMenuBar=^TMenuBar;
	TMenuBar=object(TPanel)
	private
		CreatedBy: PMenuPopup;
		OldFocus: PView;
		function MenuInFocus: Boolean;
	public
		Items: PMenuList;
		Selected: PMenuPopup;
		constructor Init(AOwner: PViewGroup);
		procedure CreateChilds; virtual;
		procedure ClosePopups;
		procedure HandleEvent(var Event: TEvent); virtual;
		function AddPopup(const Text: string; Bitmap: PBitmap; AEnabled: Boolean): PMenuPopup;
		function AddItem(const Text: string; Bitmap: PBitmap; AEnabled: Boolean; ACommand: Integer): PMenuItem;
		procedure Draw; virtual;
	end;
	TMenu=object(TMenuBar)
		constructor Init(AOwner: PViewGroup);
		procedure Show; virtual;
		procedure Draw; virtual;
		destructor Done; virtual;
	end;

implementation

uses
	GUIUtils,Surfaces;

procedure TGenericMenuItem.Setup(AOwner: PListView);
begin
	inherited Setup(AOwner);
	if AOwner^.Direction=diHorizontal then
		Size:=Width
	else
		Size:=Height;
end;

function TGenericMenuItem.Height: Integer;
var
	H: Integer;
begin
	if not (Assigned(Caption) or Assigned(Icon)) then
		H:=4
	else begin
		H:=Owner^.Canvas^.TextHeight(' ')+3;
		if Assigned(Icon) then
			H:=Maximum(H,Icon^.Height+2);
	end;
	Height:=H;
end;

function TGenericMenuItem.Width: Integer;
var
	W: Integer;
begin
	if Assigned(Icon) then
		W:=Icon^.Width+1
	else
		W:=0;
	if Assigned(Caption) then
		Inc(W,Owner^.Canvas^.TextWidth(Caption^));
	Width:=W+14;
end;

procedure TGenericMenuItem.SetCaption(const ACaption: string);
begin
	DisposeStr(Caption);
	Caption:=NewStr(ACaption);
end;

procedure TGenericMenuItem.Draw(Active,Focus: Boolean);
const
	Space=2;
begin
	with Canvas^ do begin
		if not (Assigned(Caption) or Assigned(Icon)) then begin
			SetColor(clLightGray);
			Clear;
			SetColor(clDarkGray);
			HLine(Space,1,Width-Succ(Space));
			SetColor(clWhite);
			HLine(Space,2,Width-Succ(Space));
		end else begin
			if Active and Enabled then
				SetColor(clBlue)
			else
				SetColor(clLightGray);
			Clear;
			if not Enabled then
				SetColor(clMedium)
			else if Active then
				SetColor(clWhite)
			else
				SetColor(clBlack);
			if Assigned(Icon) then begin
				GotoXY(6+Icon^.Width+Ord(Owner^.Direction=diHorizontal)*2,2);
				PutImage(Icon,2,2);
			end else
				GotoXY(5+Ord(Owner^.Direction=diHorizontal)*2,2);
			if Assigned(Caption) then
				WriteString(Caption^);
		end;
	end;
end;

procedure TGenericMenuItem.Action(Pressed: Boolean);
begin
	Abstract;
end;

procedure TGenericMenuItem.Close;
begin
end;

destructor TGenericMenuItem.Done;
begin
	inherited Done;
	DisposeStr(Caption);
end;

constructor TMenuBar.Init(AOwner: PViewGroup);
begin
	inherited Init(AOwner);
	AlwaysOnTop:=True;
	with Items^ do begin
		Direction:=diHorizontal;
		AcceptFocus:=True;
		Move(8,0);
		Show;
	end;
end;

procedure TMenuBar.CreateChilds;
begin
	if not Assigned(Items) then
		New(Items,Init(@self));
end;

procedure TMenuBar.ClosePopups;
var
	I: Integer;
begin
	for I:=0 to Items^.Count-1 do
		PGenericMenuItem(Items^.GetItem(I))^.Close;
	Selected:=nil;
end;

function TMenuBar.MenuInFocus: Boolean;
var
	Temp: PMenuBar;
begin
	Temp:=@self;
	while Assigned(Temp) and (not Temp^.Items^.Focused) do
		if Assigned(Temp^.Selected) then
			Temp:=Temp^.Selected^.Menu
		else
			Temp:=nil;
	MenuInFocus:=Assigned(Temp);
end;

function TMenuBar.AddPopup(const Text: string; Bitmap: PBitmap; AEnabled: Boolean): PMenuPopup;
var
	Act: PMenuPopup;
begin
	New(Act,Init);
	with Act^ do begin
		SetCaption(Text);
		Icon:=Bitmap;
		Enabled:=AEnabled;
	end;
	with Items^ do begin
		InsertItem(Act,Items^.Count);
		if Direction=diHorizontal then
			Resize(Width+Act^.Width,Act^.Height);
	end;
	AddPopup:=Act;
end;

function TMenuBar.AddItem(const Text: string; Bitmap: PBitmap; AEnabled: Boolean; ACommand: Integer): PMenuItem;
var
	Act: PMenuItem;
begin
	New(Act,Init);
	with Act^ do begin
		SetCaption(Text);
		Icon:=Bitmap;
		Enabled:=AEnabled;
		Command:=ACommand;
	end;
	with Items^ do begin
		InsertItem(Act,Items^.Count);
		if Direction=diHorizontal then
			Resize(Width+Act^.Width,Act^.Height);
	end;
	AddItem:=Act;
end;

procedure TMenuBar.Draw;
begin
	with Canvas^ do begin
		BeginUpdate;
		SetColor(clBlack);
		HLine(0,Height-1,Width-1);
		SetColor(clLightGray);
		Bar(0,0,Width-1,Height-2);
		EndUpdate;
	end;
end;

procedure TMenuBar.HandleEvent(var Event: TEvent);
begin
	with Event do
		case What of
		evCommand,evBroadcast: case Command of
			cmFocusChanged: if not MenuInFocus then begin
					ClosePopups;
					Items^.Select(nil);
				end;
			end;
		end;
	inherited HandleEvent(Event);
end;

procedure TMenuList.HandleEvent(var Event: TEvent);
var
	I: Integer;
	Act: PGenericMenuItem;
	Temp: PView;
begin
	with Event do
		case What of
		evMouseDown,evMouseUp,evMouseMove: begin
				if not Focused then
					if (What=evMouseMove) and (Direction=diHorizontal) and (not Assigned(Selected)) then
						Exit
					else begin
						if not Assigned(PMenuBar(Owner)^.OldFocus) then
							PMenuBar(Owner)^.OldFocus:=Screen^.Focus;
						GetFocus;
					end;
				if Direction=diHorizontal then
					I:=Where.X
				else
					I:=Where.Y;
				Act:=PGenericMenuItem(GetItemByPos(I));
				Select(Act);
				with PMenuBar(Owner)^ do
					if Assigned(Selected) and (PGenericMenuItem(Selected)<>Act) then
						ClosePopups;
				if Assigned(Act) and Act^.Enabled then
					Act^.Action(What=evMouseUp);
				EventHandled(Event);
			end;
		evCommand,evBroadcast: case Command of
			cmMouseLeave: begin
					Select(PMenuBar(Owner)^.Selected);
					EventHandled(Event);
				end;
			cmFocusChanged: if Focused and (not Assigned(PMenuBar(Owner)^.OldFocus)) then
					PMenuBar(Owner)^.OldFocus:=Screen^.Focus;
			end;
		end;
	if Event.What<>evNothing then
		inherited HandleEvent(Event);
end;

procedure TMenuItem.Action(Pressed: Boolean);
var
	Temp: PMenuBar;
begin
	if Pressed then begin
		Temp:=PMenuBar(Owner^.Owner);
		while Assigned(Temp^.CreatedBy) do
			Temp:=PMenuBar(Temp^.CreatedBy^.Owner^.Owner);
		Temp^.ClosePopups;
		Screen^.SetFocus(Temp^.OldFocus);
		if Temp^.MenuInFocus then
			Temp^.DropFocus(True);
		Temp^.OldFocus:=nil;
		Application^.Send.CommandPtr(Temp^.Handle,Command,@self);
	end;
end;

constructor TMenuPopup.Init;
begin
	inherited Init;
	New(Menu,Init(Screen));
	Menu^.CreatedBy:=@self;
end;

procedure TMenuPopup.Action(Pressed: Boolean);
var
	I,AddX,AddY: Integer;
begin
	if Assigned(Menu) and (not Menu^.Visible) then begin
		I:=Owner^.PosOf(@self);
		if Owner^.Direction=diHorizontal then begin
			AddX:=I;
			AddY:=Owner^.Height;
		end else begin
			AddX:=Owner^.Width-1;
			AddY:=I-1;
		end;
		Inc(AddX,Owner^.Canvas^.Offset.X-Menu^.Owner^.Canvas^.Offset.X);
		Inc(AddY,Owner^.Canvas^.Offset.Y-Menu^.Owner^.Canvas^.Offset.Y);
		Menu^.Move(AddX,AddY);
		Menu^.Show;
		PMenuBar(Owner^.Owner)^.Selected:=@self;
		Menu^.Items^.GetFocus;
	end;
end;

procedure TMenuPopup.Close;
begin
	if Assigned(Menu) and (Menu^.Visible) then begin
		Menu^.ClosePopups;
		Menu^.Hide;
	end;
end;

procedure TMenuPopup.Draw(Active,Focus: Boolean);
begin
	inherited Draw(Active,Focus);
	if Owner^.Direction=diVertical then
		with Canvas^ do begin
			VLine(Width-5,Height shr 1-1,Height shr 1+1);
			PutPixel(Width-4,Height shr 1);
		end;
end;

destructor TMenuPopup.Done;
begin
	inherited Done;
	if Assigned(Menu) then
		Menu^.Free;
end;

constructor TMenu.Init(AOwner: PViewGroup);
begin
	inherited Init(AOwner);
	Resize(3,3);
	with Items^ do begin
		Direction:=diVertical;
		Move(1,1);
		Resize(1,1);
		ResizeFlags:=[rfRight,rfBottom];
	end;
end;

procedure TMenu.Draw;
begin
	Canvas^.Box3D(0,0,Width-1,Height-1,clWhite,clLightGray,clDarkGray,1);
end;

procedure TMenu.Show;
var
	I,W,H: Integer;
	Act: PGenericMenuItem;
begin
	W:=0;
	H:=0;
	for I:=0 to Items^.Count-1 do begin
		Act:=PGenericMenuItem(Items^.GetItem(I));
		W:=Maximum(W,Act^.Width);
		Inc(H,Act^.Size);
	end;
	if (W>0) and (H>0) then begin
		Resize(W+2,H+2);
		MoveToTop;
		inherited Show;
	end;
end;

destructor TMenu.Done;
begin
	if Assigned(CreatedBy) then
		CreatedBy^.Menu:=nil;
	inherited Done;
end;

end.